(** Copyright (c) 2018 by Swirlds, Inc.
   Implemented by Karl Crary.
*)

Require Import Nat.
Require Import Bool.
Require Import Omega.
Require Import List.
Require Import Sorting.
Require Import Permutation.

Require Import Tact.
Require Import Decide.
Require Import Relation.
Require Import Cardinality.
Require Import Calculate.
Require Import Majority.


Module NatSort := Sort NatOrder.

Definition median (l : list nat) : nat :=
  let l' := NatSort.sort l
  in let len := length l'
  in
    nth (pred (S len / 2)) l' 0.


Lemma sort_unique :
  forall (T : Type) (R : T -> T -> Prop),
    order T R
    -> (forall x y, R x y \/ R y x)
    -> (forall x y, decidable (R x y))
    -> forall (l m : list T),
         Sorted R l
         -> Sorted R m
         -> Permutation l m
         -> l = m.
Proof.
intros T R Hord Htotal Hdec l m Hl Hm Hlm.
revert m Hl Hm Hlm.
induct l.

(** nil *)
{
intros m _ _ Hperm.
symmetry.
apply Permutation_nil; auto.
}

(** cons *)
{
intros x l' IH m Hl Hm Hlm.
destruct m as [| y m'].
  {
  exfalso.
  symmetry in Hlm.
  eapply Permutation_nil_cons; eauto.
  }
so (Sorted_inv _#4 Hl) as (Hl' & _).
so (Sorted_inv _#4 Hm) as (Hm' & _).
assert (x = y).
  {
  destruct (Permutation_in _ Hlm (or_introl (eq_refl _))) as [| Hinx].
    {
    subst y.
    reflexivity.
    }
  destruct (Permutation_in _ (Permutation_sym Hlm) (or_introl (eq_refl _))) as [| Hiny].
    {
    subst y.
    reflexivity.
    }
  apply (ord_antisym T R); auto.
    {
    exact (Forall_forall _#3 andel (@Sorted_extends _ _ (ord_trans _ _ Hord) _ _ Hl) _ Hiny).
    }

    {
    exact (Forall_forall _#3 andel (@Sorted_extends _ _ (ord_trans _ _ Hord) _ _ Hm) _ Hinx).
    }
  }
subst y.
f_equal.
apply IH; auto.
eapply Permutation_cons_inv; eauto.
}
Qed.


Lemma sorted_impl :
  forall T R R' l,
    inclusion T R R'
    -> Sorted R l
    -> Sorted R' l.
Proof.
intros T R R' l Hincl Hl.
apply Sorted_LocallySorted_iff.
so (Sorted_LocallySorted_iff _#3 andel Hl) as H.
clear Hl.
induct H; eauto using LocallySorted.
Qed.


Lemma nat_sort_sorted :
  forall l,
    Sorted le (NatSort.sort l).
Proof.
intro l.
(** The lemmas Sorted_sort and LocallySorted_sort are reversed in the library. *)
refine (sorted_impl _#4 _ (NatSort.LocallySorted_sort _)).
intros x y Hxy.
apply leb_complete; auto.
Qed.


Lemma nat_sort_unique :
  forall (l m : list nat),
    Permutation l m
    -> NatSort.sort l = NatSort.sort m.
Proof.
intros l m Hperm.
assert (inclusion _ (fun x y => is_true (Nat.leb x y)) le) as Hincl.
  {
  intros x y Hxy.
  apply leb_complete; auto.
  }
apply (sort_unique _ le).
  {
  apply Build_order.
    { unfold reflexive; intros; omega. }

    { unfold transitive; intros; omega. }

    { unfold antisymmetric; intros; omega. }
  }

  {
  intros x y.
  omega.
  }

  {
  exact Nat.le_decidable.
  }

  {
  apply nat_sort_sorted.
  }

  {
  apply nat_sort_sorted.
  }

  {
  apply (@perm_trans _ _ l).
    {
    apply Permutation_sym.
    apply NatSort.Permuted_sort.
    }
  eapply perm_trans; eauto.
  apply NatSort.Permuted_sort.
  }
Qed.


Lemma median_unique :
  forall (l m : list nat),
    Permutation l m
    -> median l = median m.
Proof.
intros l m Hlm.
unfold median.
rewrite -> (nat_sort_unique _ _ Hlm).
reflexivity.
Qed.


Lemma Permutation_of_map :
  forall (A B : Type) (f : A -> B) (l : list A) (m : list B),
    Permutation (map f l) m
    -> exists m',
         m = map f m'
         /\ Permutation l m'.
Proof.
intros A B f l m Hperm.
remember (map f l) as n eqn:Heqn.
revert l Heqn.
induct Hperm.

(** nil *)
{
intros l Heq.
destruct l as [| x l].
  2:{ discriminate Heq. }
exists nil.
auto.
}

(** skip *)
{
intros x n m _ IH l Heq.
destruct l as [| y l].
  { discriminate Heq. }
cbn in Heq.
injection Heq.
intros -> ->.
so (IH l (eq_refl _)) as (p & -> & Hperm).
exists (cons y p).
cbn.
auto.
}

(** swap *)
{
intros x y n l Heq.
destruct l as [| u l].
  { discriminate Heq. }
destruct l as [| v l].
  { discriminate Heq. }
cbn in Heq.
injection Heq.
intros -> -> ->.
exists (cons v (cons u l)).
cbn.
split; auto.
apply perm_swap.
}

(** trans *)
{
intros n m p _ IH1 _ IH2 l ->.
so (IH1 l (eq_refl _)) as (q & -> & Hperm1).
so (IH2 q (eq_refl _)) as (r & -> & Hperm2).
exists r.
split; auto.
eapply perm_trans; eauto.
}
Qed.


Lemma HdRel_app_inv :
  forall (T : Type) (R : T -> T -> Prop) (l l' : list T) (x : T),
    HdRel R x (l ++ l')
    -> HdRel R x l.
Proof.
intros T R l1 l2 x Hr.
destruct l1 as [| y l1].
  {
  apply HdRel_nil.
  }

  {
  apply HdRel_cons.
  eapply HdRel_inv; eauto.
  }
Qed.      


Lemma Sorted_app_inv :
  forall (T : Type) (R : T -> T -> Prop) (l l' : list T),
    Sorted R (l ++ l') 
    -> Sorted R l /\ Sorted R l'.
Proof.
intros T R l1 l2 Hsorted.
remember (l1 ++ l2) as l eqn:Heql.
revert l Heql Hsorted.
induct l1.

(** nil *)
{
intros ? -> Hsorted.
split; auto.
}

(** cons *)
{
intros x l1 IH ? -> Hsorted.
cbn in Hsorted.
invert Hsorted.
intros Hsorted' Hr.
so (IH _ (eq_refl _) Hsorted') as (H1 & H2).
split; auto.
eapply Sorted_cons; eauto.
eapply HdRel_app_inv; eauto.
}
Qed.


Lemma NoDup_app_inv :
  forall (T : Type) (l l' : list T),
    NoDup (l ++ l')
    -> NoDup l /\ NoDup l'.
Proof.
intros T l1 l2 Hdist.
remember (l1 ++ l2) as l eqn:Heql.
revert l Heql Hdist.
induct l1.

(** nil *)
{
intros ? -> Hdist.
split.
  {
  apply NoDup_nil.
  }

  {
  exact Hdist.
  }
}

(** cons *)
{
intros x l1 IH ? -> Hdist.
cbn in Hdist.
invert Hdist.
intros Hnin Hdist'.
so (IH _ (eq_refl _) Hdist') as (Hdist1 & Hdist2).
split; auto.
apply NoDup_cons; auto.
contradict Hnin.
apply in_or_app.
left; auto.
}
Qed.


Lemma Sorted_cons_inv :
  forall (T : Type) (R : T -> T -> Prop) (x y : T) (l : list T),
    reflexive T R
    -> transitive T R
    -> Sorted R (cons x l)
    -> In y (cons x l)
    -> R x y.
Proof.
intros T R x y l Hrefl Htrans Hsorted Hin.
destruct Hin as [| Hin].
  {
  subst y.
  apply Hrefl.
  }

  {
  invert (@Sorted_StronglySorted _ _ Htrans _ Hsorted).
  intros _ Hord.
  exact (Forall_forall _#3 andel Hord y Hin).
  }
Qed.


Lemma Sorted_snoc_inv :
  forall (T : Type) (R : T -> T -> Prop) (x y : T) (l : list T),
    reflexive T R
    -> transitive T R
    -> Sorted R (l ++ x :: nil)
    -> In y (l ++ x :: nil)
    -> R y x.
Proof.
intros T R x y l Hrefl Htrans Hsorted Hin.
so (in_app_or _#4 Hin) as [Hin' | Hin'].
  {
  so (in_split _#3 Hin') as (l1 & l2 & ->).
  rewrite <- app_assoc in Hsorted.
  rewrite <- app_comm_cons in Hsorted.
  so (Sorted_app_inv _#4 Hsorted) as (_ & Hsorted').
  invert (@Sorted_StronglySorted _ _ Htrans _ Hsorted').
  intros _ Hord.
  refine (Forall_forall _#3 andel Hord x _).
  apply in_or_app; right; left; auto.
  }

  {
  destruct Hin' as [| H].
    {
    subst y.
    apply Hrefl.
    }

    {
    destruct H.
    }
  }
Qed.


Lemma nil_decide :
  forall T (l : list T), decidable (l = nil).
Proof.
intros T l.
destruct l.
  { left; auto. }
right.
intros H.
discriminate H.
Qed.


Definition middle (l : list nat) : nat :=
  nth (pred (S (length l) / 2)) l 0.


Lemma middle_In :
  forall l,
    l <> nil
    -> In (middle l) l.
Proof.
intros l Hnnil.
unfold middle.
apply nth_In.
assert (length l > 0) as Hgt.
  {
  destruct l; cbn; try omega.
  destruct Hnnil; auto.
  }
remember (length l) as n eqn:H.
clear l H Hnnil.
so (Nat.div_mod (S n) 2 (Nat.neq_succ_0 _)).
omega.
Qed.


Lemma middle_center :
  forall x l y,
    l <> nil
    -> middle (x :: l ++ y :: nil) = middle l.
Proof.
intros x l y Hnnil.
unfold middle.
cbn [length].
rewrite -> app_length.
cbn [length].
replace (S (length l + 1)) with (S (S (length l))) by omega.
replace (S (S (S (length l))) / 2) with (S (S (length l) / 2)).
2:{
  symmetry.
  apply one_half_plus_two.
  }
cbn [pred].
assert (length l > 0) as Hnonzero.
  {
  destruct l; cbn; try omega.
  destruct Hnnil; auto.
  }
remember (length l) as n eqn:Heqn.
assert (S n / 2 > 0) as Hgt.
  {
  destruct n; try omega.
  replace (S (S n)) with (2 + n) by omega.
  rewrite -> one_half_plus_two.
  omega.
  }
assert (S n / 2 <= n) as Hle.
  {
  so (Nat.div_mod (S n) 2 (Nat.neq_succ_0 _)).
  omega.
  }
remember (S n / 2) as m eqn:Heqm.
destruct m as [| m].
  {
  omega.
  }
cbn [pred].
cbn [nth].
apply app_nth1.
omega.
Qed.


Lemma majority_median :
  forall (T : Type) (P : T -> Prop) (f : T -> nat) (l : list T),
    (forall (x y : T), decidable (x = y))
    -> NoDup l
    -> majority P (fun x => In x l)
    -> exists x y,
         P x
         /\ P y
         /\ f x <= median (map f l) <= f y.
Proof.
intros T P f l Hdec Hdistl Hmaj.
destruct Hmaj as (m & n & Hincl & HcardP & Hcardl & Hgt).
so (cardinality_fun _#4 Hcardl (cardinality_In _ _ Hdistl)) as Heqn.
clear Hcardl.
destruct HcardP as (p & Hdistp & Hp & <-).
so (Permutation_of_map _#5 (NatSort.Permuted_sort (map f l))) as (l' & Heq & Hperm).
so (nat_sort_sorted (map f l)) as Hsorted.
replace (median (map f l)) with (middle (NatSort.sort (map f l))) by reflexivity.
rewrite -> Heq in Hsorted |- *.
cut (exists x y, In x p /\ In y p /\ f x <= middle (map f l') <= f y).
  {
  intros (x & y & Hx & Hy & Hrange).
  exists x, y.
  do2 2 split; auto.
    {
    apply Hp; auto.
    }

    {
    apply Hp; auto.
    }
  }
assert (List.incl p l') as Hpl'.
  {
  intros x Hx.
  eapply Permutation_in; eauto.
  apply Hincl.
  apply Hp; auto.
  }
so (Permutation_NoDup _#3 Hperm Hdistl) as Hdistl'.
rewrite -> (Permutation_length _#3 Hperm) in Heqn.
clear l Heq Hperm Hdistl Hp Hincl.
revert p l' Hdistp Hdistl' Heqn Hpl' Hsorted Hgt.
wfinduct n using lt_wf.
intros n IH p l Hdistp Hdistl Heqn Hpl Hsorted Hgt.
destruct l as [| x l].
  {
  subst n.
  destruct p as [| x p].
    {
    cbn in Hgt.
    omega.
    }
    
    {
    exploit (Hpl x) as H.
      {
      left; auto.
      }
    destruct H.
    }
  }
so (nil_decide _ l) as [-> | Hnnil].
  {
  assert (In x p) as Hpx.
    {
    destruct p as [| y p'].
      {
      cbn in Hgt.
      omega.
      }
      
      {
      left.
      destruct (Hpl y (or_introl (eq_refl _))) as [| H]; auto.
      destruct H.
      }
    }
  exists x, x.
  auto.
  }
assert (forall (p1 p2 : list T) u v, In v (p1 ++ p2) -> In v (p1 ++ u :: p2)) as Hmidimp.
  {
  intros p1 p2 u v H.
  so (in_app_or _#4 H) as [H' | H'].
    {
    apply in_or_app; left; auto.
    }

    {
    apply in_or_app; right; right; auto.
    }
  }
so (exists_last Hnnil) as (l' & y & ->).
clear Hnnil.
assert (map f l' = nil -> l' = nil) as Hnilimp.
  {
  intro Heq.
  destruct l' as [|]; auto.
  discriminate Heq.
  }
so (dec_In _ x p Hdec) as [Hpx | Hnpx]; so (dec_In _ y p Hdec) as [Hpy | Hnpy].
  (** both in *)
  {
  exists x, y.
  do2 2 split; auto.
  exploit (middle_In (map f (x :: l' ++ y :: nil))) as Hin.
    {
    cbn.
    intro H.
    discriminate H.
    }
  remember (middle (map f (x :: l' ++ y :: nil))) as u eqn:Hequ.
  split.
    {
    cbn [map] in Hsorted, Hin.
    eapply Sorted_cons_inv; eauto.
      {
      unfold reflexive; intros; omega.
      }

      {
      unfold transitive; intros; omega.
      }
    }

    {
    cbn [map] in Hsorted, Hin.
    rewrite -> map_app in Hsorted, Hin.
    cbn [map] in Hsorted, Hin.
    rewrite -> app_comm_cons in Hin.
    eapply Sorted_snoc_inv; eauto.
      {
      unfold reflexive; intros; omega.
      }

      {
      unfold transitive; intros; omega.
      }

      {
      rewrite <- app_comm_cons; auto.
      }
    }
  }

  (** left in, right out *)
  {
  assert (length l' < n) as Hltn.
    {
    cbn in Heqn.
    rewrite -> app_length in Heqn.
    omega.
    }
  so (in_split _#3 Hpx) as (p1 & p2 & ->).
  exploit (IH (length l') Hltn (p1 ++ p2) l') as H; auto.
    {
    eapply NoDup_remove_1; eauto.
    }

    {
    invert Hdistl.
    intros _ Hdistl'.
    so (NoDup_app_inv _#3 Hdistl') as (Hdistl'' & _).
    exact Hdistl''.
    }

    {
    so (NoDup_remove_2 _#4 Hdistp) as Hnin.
    cut (forall z, In z (p1 ++ x :: p2) -> z <> x -> In z l').
      {
      intro Hprop.
      apply List.incl_app.
        {
        intros z Hz.
        apply Hprop.
          {
          apply in_or_app; left; auto.
          }

          {
          intro; subst z.
          destruct Hnin.
          apply in_or_app; left; auto.
          }
        }

        {
        intros z Hz.
        apply Hprop.
          {
          apply in_or_app; right; right; auto.
          }

          {
          intro; subst z.
          destruct Hnin.
          apply in_or_app; right; auto.
          }
        }
      }
    intros z Hz Hneq.
    so (Hpl _ Hz) as Hin.
    destruct Hin as [Heq | Hin].
      {
      subst z.
      destruct Hneq; auto.
      }
    so (in_app_or _#4 Hin) as [Hin' | Hin']; auto.
    destruct Hin' as [| H].
      {
      subst z.
      contradiction.
      }
      
      { destruct H. }
    }

    {
    invert Hsorted.
    cbn in Hsorted.
    intros Hsorted' _.
    rewrite -> map_app in Hsorted'.
    so (Sorted_app_inv _#4 Hsorted') as (H & _).
    exact H.
    }

    {
    subst n.
    rewrite -> app_length.
    cbn [length] in Hgt.
    rewrite -> !app_length in Hgt.
    cbn [length] in Hgt.
    replace (S (length l' + 1)) with (2 + length l') in Hgt by omega.
    unfold one_half in *.
    rewrite -> one_half_plus_two in Hgt.
    omega.
    }

    {
    destruct H as (u & v & Hu & Hv & Hrange).
    exists u, v.
    do2 2 split.
      {
      so (in_app_or _#4 Hu) as [Hu' | Hu'].
        {
        apply in_or_app; left; auto.
        }

        {
        apply in_or_app; right; right; auto.
        }
      }

      {
      so (in_app_or _#4 Hv) as [Hv' | Hv'].
        {
        apply in_or_app; left; auto.
        }

        {
        apply in_or_app; right; right; auto.
        }
      }

      {
      cbn [map].
      rewrite -> map_app.
      cbn [map].
      rewrite -> middle_center; auto.
      intro Hnil.
      so (Hnilimp Hnil); subst l'.
      cbn [app] in Hpl.
      so (Hpl _ (Hmidimp _#4 Hu)) as [|[|H]].
        {
        subst u.
        so (NoDup_remove_2 _#4 Hdistp).
        contradiction.
        }

        {
        subst u.
        destruct Hnpy.
        apply Hmidimp; auto.
        }

        { destruct H. }
      }
    }
  }

  (** left out, right in *)
  {
  assert (length l' < n) as Hltn.
    {
    cbn in Heqn.
    rewrite -> app_length in Heqn.
    omega.
    }
  so (in_split _#3 Hpy) as (p1 & p2 & ->); clear Hpy.
  exploit (IH (length l') Hltn (p1 ++ p2) l') as H; auto.
    {
    eapply NoDup_remove_1; eauto.
    }

    {
    invert Hdistl.
    intros _ Hdistl'.
    so (NoDup_app_inv _#3 Hdistl') as (Hdistl'' & _).
    exact Hdistl''.
    }

    {
    so (NoDup_remove_2 _#4 Hdistp) as Hnin.
    cut (forall z, In z (p1 ++ y :: p2) -> z <> y -> In z l').
      {
      intro Hprop.
      apply List.incl_app.
        {
        intros z Hz.
        apply Hprop.
          {
          apply in_or_app; left; auto.
          }

          {
          intro; subst z.
          destruct Hnin.
          apply in_or_app; left; auto.
          }
        }

        {
        intros z Hz.
        apply Hprop.
          {
          apply in_or_app; right; right; auto.
          }

          {
          intro; subst z.
          destruct Hnin.
          apply in_or_app; right; auto.
          }
        }
      }
    intros z Hz Hneq.
    so (Hpl _ Hz) as Hin.
    destruct Hin as [Heq | Hin].
      {
      subst z.
      contradiction.
      }
    so (in_app_or _#4 Hin) as [Hin' | Hin']; auto.
    destruct Hin' as [| H].
      {
      subst z.
      destruct Hneq; auto.
      }
      
      { destruct H. }
    }

    {
    invert Hsorted.
    cbn in Hsorted.
    intros Hsorted' _.
    rewrite -> map_app in Hsorted'.
    so (Sorted_app_inv _#4 Hsorted') as (H & _).
    exact H.
    }

    {
    subst n.
    rewrite -> app_length.
    cbn [length] in Hgt.
    rewrite -> !app_length in Hgt.
    cbn [length] in Hgt.
    replace (S (length l' + 1)) with (2 + length l') in Hgt by omega.
    unfold one_half in *.
    rewrite -> one_half_plus_two in Hgt.
    omega.
    }

    {
    destruct H as (u & v & Hu & Hv & Hrange).
    exists u, v.
    do2 2 split.
      {
      so (in_app_or _#4 Hu) as [Hu' | Hu'].
        {
        apply in_or_app; left; auto.
        }

        {
        apply in_or_app; right; right; auto.
        }
      }

      {
      so (in_app_or _#4 Hv) as [Hv' | Hv'].
        {
        apply in_or_app; left; auto.
        }

        {
        apply in_or_app; right; right; auto.
        }
      }

      {
      cbn [map].
      rewrite -> map_app.
      cbn [map].
      rewrite -> middle_center; auto.
      intro Hnil.
      so (Hnilimp Hnil); subst l'.
      cbn [app] in Hpl.
      so (Hpl _ (Hmidimp _#4 Hu)) as [|[|H]].
        {
        subst u.
        destruct Hnpx.
        apply Hmidimp; auto.
        }

        {
        subst u.
        so (NoDup_remove_2 _#4 Hdistp).
        contradiction.
        }

        { destruct H. }
      }
    }
  }

  (** both out *)
  {
  assert (length l' < n) as Hltn.
    {
    cbn in Heqn.
    rewrite -> app_length in Heqn.
    omega.
    }
  exploit (IH (length l') Hltn p l') as H; auto.
    {
    invert Hdistl.
    intros _ Hdistl'.
    so (NoDup_app_inv _#3 Hdistl') as (Hdistl'' & _).
    exact Hdistl''.
    }

    {
    intros z Hz.
    so (Hpl _ Hz) as Hin.
    destruct Hin as [Heq | Hin].
      {
      subst z.
      contradiction.
      }
    so (in_app_or _#4 Hin) as [Hin' | Hin']; auto.
    destruct Hin' as [| H].
      {
      subst z.
      contradiction.
      }
      
      { destruct H. }
    }

    {
    invert Hsorted.
    cbn in Hsorted.
    intros Hsorted' _.
    rewrite -> map_app in Hsorted'.
    so (Sorted_app_inv _#4 Hsorted') as (H & _).
    exact H.
    }

    {
    subst n.
    cbn [length] in Hgt.
    rewrite -> !app_length in Hgt.
    cbn [length] in Hgt.
    replace (S (length l' + 1)) with (2 + length l') in Hgt by omega.
    unfold one_half in *.
    rewrite -> one_half_plus_two in Hgt.
    omega.
    }

    {
    destruct H as (u & v & Hu & Hv & Hrange).
    exists u, v.
    do2 2 split; auto.
    cbn [map].
    rewrite -> map_app.
    cbn [map].
    rewrite -> middle_center; auto.
    intro Hnil.
    so (Hnilimp Hnil); subst l'.
    cbn [app] in Hpl.
    so (Hpl _ Hu) as [|[|H]].
      {
      subst u.
      contradiction.
      }

      {
      subst u.
      contradiction.
      }

      { destruct H. }
    }
  }
Qed.
